Require Import List.
Require Import Arith.
Require Import Program.Tactics.

(** Finite maps and lemmas **)

Definition PartialMap (a b : Type) := a -> option b.

Notation "x |-> y" := (PartialMap x y) (at level 70, no associativity).

(** Internal model **)

Variable F : Type.
Variable eq_f_dec : forall (x y : F), {x=y}+{x<>y}.
Variable C : Type.
Variable epsilon : C.

Definition M := (F |-> C).

Definition beq_f (x y : F) : bool :=
  match eq_f_dec x y with
    | left _ => true
    | right _ => false
  end.

Lemma beq_f_refl (x : F) : beq_f x x = true.
  unfold beq_f; destruct (eq_f_dec x x); tauto.
  Qed.  

Lemma beq_f_true (x y : F) : beq_f x y = true -> x = y.
  Proof.
    unfold beq_f; destruct (eq_f_dec x y); [ tauto | intros; discriminates].
  Qed.

(** Lemmas about finite maps **)

Definition update (f : F) (c : option C) (m : M) : M :=
  fun f' => if beq_f f f' then c else m f'.

Definition inDom {a b : Type} (f : a |-> b) (x : a) : Prop :=
  exists (y : b), f x = Some y.

Definition mapsTo (m : M) (f : F) (c : C) : Prop := 
  m f = Some c.

Definition notIn (m : M) (f : F) : Prop :=
  m f = None.

Hint Unfold inDom mapsTo notIn mapsTo.

Lemma lookupUpdated (f : F) (c : option C) (m : M) :
  (update f c m) f = c.
  Proof.
    unfold update; rewrite beq_f_refl; reflexivity.
  Qed.

Lemma lookupNUpdated (f g : F) (c : option C) (m : M) :
  f <> g -> update f c m g = m g.
  Proof.
    unfold update; remember (beq_f f g) as b; destruct b; intros F.
    - exfalso; apply F; now apply beq_f_true.
    - reflexivity.
  Qed.

Hint Resolve lookupUpdated lookupNUpdated.

(* Boring lemmas about partial maps *)

Definition applyNone : option C := None.

Hint Resolve applyNone.

Lemma inDomUpdate : forall m f g c, f <> g -> inDom m g -> inDom (update f c m) g.
  Proof.
    intros m f g c F [y H]; autounfold; rewrite lookupNUpdated; eauto.
  Qed.    

Lemma inDomUpdate' : forall m f g c, f <> g -> inDom (update f c m) g -> inDom m g.
  Proof.
    intros m f g c F [y H]; exists y; rewrite lookupNUpdated in H; eauto.
  Qed.    

Lemma inDomUpdate'' : forall m f c, inDom (update f (Some c) m) f.
  Proof.
    intros; exists c; rewrite lookupUpdated; reflexivity.
  Qed.

Lemma inDomUpdate''' : forall f g m, inDom (update f None m) g -> inDom m g.
  Proof.
    intros f g m [y H]; destruct (eq_f_dec f g); subst.
    - exfalso; rewrite lookupUpdated in *; discriminate.
    - rewrite lookupNUpdated in H; eauto.
  Qed.

Lemma notInUpdate : forall m f g c, f <> g -> notIn m g -> notIn (update f c m) g.
  Proof.
    intros m f g c F; autounfold; destruct (eq_f_dec f g).
      - contradiction; tauto.
      - rewrite lookupNUpdated; eauto.
  Qed.    

Lemma notInUpdate' : forall m f g c, f <> g -> notIn (update f c m) g -> notIn m g.
  Proof.
    autounfold; intros m f g c F H; rewrite lookupNUpdated in H; eauto.
  Qed.    

Lemma notInUpdate'' : forall m f g, f <> g -> notIn m g -> notIn (update f None m) g.
  Proof.
    intros; apply notInUpdate; assumption.
  Qed.

Lemma notInUpdate''' : forall m f g, f <> g -> notIn (update f None m) g -> notIn m g.
  Proof.
    intros; eapply notInUpdate'; eauto.
  Qed.

Hint Resolve inDomUpdate inDomUpdate' inDomUpdate'' inDomUpdate'''
  notInUpdate notInUpdate' notInUpdate'' notInUpdate'''.

Lemma elemContradiction : forall m f, inDom m f -> notIn m f -> False.
  Proof.
    autounfold; intros m f [y E] N; rewrite E in N; discriminate.
  Qed.

Hint Resolve elemContradiction.

(** Separation logic predicates *)

Definition disjoint (m1 m2 : M) : Prop :=
  forall (f : F), (inDom m1 f -> notIn m2 f)
    /\ (inDom m2 f -> notIn m1 f).

Definition union (m1 m2 : M) (H : disjoint m1 m2) : M :=
  fun x =>
    match m1 x with
        | None => m2 x
        | Some y => Some y
    end.

Definition splits (m1 m2 m : M) :=
  { H : disjoint m1 m2 & forall f, union m1 m2 H f = m f}.

(** Predicates **)

Inductive Predicate : Type :=
  | MAPSTO : F -> C -> Predicate
  | EMPTY: Predicate
  | STAR : Predicate -> Predicate -> Predicate.

Fixpoint holds (p : Predicate) (m : M) : Prop :=
  match p with
    | STAR p1 p2 => exists m1 m2, splits m1 m2 m /\ holds p1 m1 /\ holds p2 m2 
    | MAPSTO f c => mapsTo m f c /\ forall f', f <> f' -> notIn m f'
    | EMPTY => forall f, notIn m f
  end.

Hint Unfold holds.

(** Commands with their denotational semantics **)

Inductive Cmd : Type :=
  | Add : F -> Cmd
  | Remove : F -> Cmd
  | Replace : F -> C -> C -> Cmd
  | Seq : Cmd -> Cmd -> Cmd.

Fixpoint semantics (cmd : Cmd) (m : M) : M :=
  match cmd with
   | Add f => update f (Some epsilon) m
   | Remove f => update f None m
   | Replace f _ d => update f (Some d) m
   | Seq c1 c2 => semantics c2 (semantics c1 m)
  end.

Hint Unfold semantics disjoint inDom union.

Fixpoint post (cmd : Cmd) : M -> Prop :=
  match cmd with
    | Add f => holds (MAPSTO f epsilon)
    | Remove f => holds (EMPTY)
    | Replace f c d => holds (MAPSTO f d)
    | Seq c1 c2 => post c2
  end.

Fixpoint pre (cmd : Cmd) : M -> Prop :=
  match cmd with
    | Add f => holds (EMPTY)
    | Remove f => holds (MAPSTO f epsilon)
    | Replace f c d => holds (MAPSTO f c)
    | Seq c1 c2 => fun m => pre c1 m /\ (forall m, post c1 m -> pre c2 m)
  end.

(* Soundness of semantics *)

Theorem soundness (cmd : Cmd) : forall m, pre cmd m -> post cmd (semantics cmd m).
  Proof.
    induction cmd; intros; simpl in *; destruct_conjs; try split; eauto.
    intros g; destruct (eq_f_dec f g); subst; eauto.
  Qed.

(* Defining mod, addr and separation *)

Inductive Modifies : F -> Cmd -> Prop :=
  | ModAdd : forall f, Modifies f (Add f)
  | ModRemove : forall f, Modifies f (Remove f)
  | ModReplace : forall f c d, Modifies f (Replace f c d)
  | ModSeq1 : forall f cmd1 cmd2, Modifies f cmd1 -> Modifies f (Seq cmd1 cmd2) 
  | ModSeq2 : forall f cmd1 cmd2, Modifies f cmd2 -> Modifies f (Seq cmd1 cmd2).

Hint Constructors Modifies.

Inductive Addr : F -> Predicate -> Prop :=
  | DomMaps : forall f c, Addr f (MAPSTO f c)
  | DomStarL : forall f P Q, Addr f P -> Addr f (STAR P Q)
  | DomStarR : forall f P Q, Addr f Q -> Addr f (STAR P Q).

Hint Constructors Addr.

Definition Separate (cmd : Cmd) (R : Predicate) : Prop :=
 forall f, Modifies f cmd -> Addr f R -> False.

Hint Unfold Separate.

(* Boring lemmas about disjointness and splitting *)

Lemma notInSplit : forall m m1 m2 f, splits m1 m2 m -> notIn m1 f -> notIn m2 f ->  notIn m f.
  Proof.
    autounfold; intros m1 m2 m f [H3 H4] H1 H2; rewrite <- H4. unfold union. rewrite H1, H2; reflexivity.
  Qed.

Hint Resolve notInSplit.

Lemma updateSeparate : forall f c m1 m2, notIn m2 f -> disjoint (update f c m1) m2 -> disjoint m1 m2.
  Proof.
  intros f c m1 m2 N D g; destruct (eq_f_dec f g) as [E | F]; [subst f | ];
    destruct (D g); split; eauto.
  Qed.

Hint Resolve updateSeparate.

Lemma disjointUpdate (m1 m2 m : M) (f : F) (c : option C) :
  disjoint m1 m2 -> notIn m2 f -> disjoint (update f c m1) m2.
  Proof.
    intros D N g; destruct (eq_f_dec f g) as [E | F]; destruct (D g); [ subst f | ]; split; eauto.
  Qed.

Hint Resolve disjointUpdate.

Lemma lookupUnion (m1 m2 m : M) (f : F) (c : option C) :
   notIn m2 f ->
   (forall H f, union m1 m2 H f = m f) ->
   forall H g, (union (update f c m1) m2 H g = update f c m g).
  Proof.
      intros N U H g; destruct (eq_f_dec f g); [subst f | ].
      - unfold union; rewrite lookupUpdated, lookupUpdated; destruct c; trivial.
      - unfold union in *; rewrite lookupNUpdated; [ | assumption]. 
        rewrite U; [ now rewrite lookupNUpdated  | now eauto].        
  Qed.

Hint Resolve lookupUnion.

Lemma splitUpdate (m1 m2 m : M) (f : F) (c : option C) :
  notIn m2 f ->
  splits m1 m2 m -> splits (update f c m1) m2 (update f c m).
  Proof.
    intros N [H1 H2]; assert (H : disjoint (update f c m1) m2) by now auto.
    exists H; eauto.
  Qed.

Hint Resolve splitUpdate.

Lemma SeparateStarL : forall cmd R1 R2, Separate cmd (STAR R1 R2) -> Separate cmd R1.
  Proof.
    intros; induction cmd; intros; eauto.
  Qed.

Lemma SeparateStarR : forall cmd R1 R2, Separate cmd (STAR R1 R2) -> Separate cmd R2.
  Proof.
    intros; induction cmd; intros; eauto.
  Qed.

Hint Resolve SeparateStarL SeparateStarR.

(* Here is the interesting part of the frame rule proof *)

Lemma splitSemantics (cmd : Cmd) :
     forall m1 m2 m R,
     (forall f, Modifies f cmd -> notIn m2 f) ->
     holds R m2 -> 
     splits m1 m2 m ->
     splits (semantics cmd m1) m2 (semantics cmd m).
  Proof.
    induction cmd; intros; autounfold; eauto.
    eapply IHcmd2; eauto.
  Qed.

Hint Resolve splitSemantics.

Lemma SeparateLemma (cmd : Cmd) (R : Predicate) : 
  Separate cmd R ->
  (forall f m, Modifies f cmd -> holds R m -> notIn m f).
  Proof.
    induction R;intros D f' m M H.
     - simpl holds in *; destruct_conjs; destruct (eq_f_dec f f') as [E | F].
       * destruct cmd; inversion M; subst; exfalso; eauto.
       * eauto.
     - eauto.
     - destruct H as [m1 [m2 H]]; destruct_conjs; eapply (notInSplit m m1 m2 _ H); eauto.
Qed.

Hint Resolve SeparateLemma.
    
Theorem frameRule (cmd : Cmd) (P Q R : Predicate)
        (H : forall m, holds P m -> holds Q (semantics cmd m)) 
        (D : Separate cmd R)
  : forall m, holds (STAR P R) m -> holds (STAR Q R) (semantics cmd m).
  Proof.
    intros m [m1 [m2 A]]; destruct_conjs; exists (semantics cmd m1); exists m2; split; eauto.
  Qed.


  