Require Import List.
Require Import Arith.
Require Import Program.Tactics.

(** Finite maps and lemmas **)

Definition PartialMap (a b : Type) := a -> option b.

(** Internal model **)

Definition I := nat.
Definition A := prod I bool.
Definition L := nat.
Definition C := sum L I.
Definition M := PartialMap A C.

Definition epsilon : C := inl 0.

Definition contents : I -> A := fun x => (x,true).

Definition next : I -> A := fun x => (x,false).

Definition line : L -> C := fun l => inl l.

Definition ptr : I -> C := fun l => inr l.

Lemma next_not_contents : forall x y,   next x <> contents y.
  intros x y F; discriminate F.
  Qed.

Hint Resolve next_not_contents.

Lemma next_inj : forall x y, x <> y -> next x <> next y.
  intros x y F1 F2; injection F2; firstorder.
  Qed.

Hint Resolve next_inj.
(** Lemmas about finite maps **)

Definition beq_a (x y : A) : bool :=
  match x,y with
    (i1,b1),(i2,b2) => andb  (negb (xorb b1 b2)) (beq_nat i1 i2)
  end.

Lemma beq_a_refl : forall a, true = beq_a a a.
  Proof.
    intros a; destruct a as [i b]; destruct b; simpl beq_a; rewrite <- beq_nat_refl; reflexivity.
  Qed.

Lemma beq_a_true : forall x y, beq_a x y = true -> x = y.
  Proof.
    intros [i1 b1] [i2 b2] H.
    destruct b1; destruct b2; discriminate || simpl beq_a in *; f_equal; now (apply beq_nat_true).
  Qed.

Lemma beq_a_false : forall x y, beq_a x y = false -> x <> y.
  Proof.
    intros [i1 b1] [i2 b2] H.
    destruct b1; destruct b2; discriminate 
      || simpl beq_a in *; intros F; injection  F; now apply beq_nat_false.
  Qed.

Lemma dec_eq_a (x y : A) : {x = y}+{x<>y}.
  repeat (decide equality).
  Qed.

Definition update (f : A) (c : option C) (m : M) : M :=
  fun f' => if beq_a f f' then c else m f'.

Definition inDom {a b : Type} (f : PartialMap a b) (x : a) : Prop :=
  exists (y : b), f x = Some y.

Definition mapsTo (m : M) (f : A) (c : C) : Prop := 
  m f = Some c.

Definition notIn (m : M) (f : A) : Prop :=
  m f = None.

Hint Unfold inDom mapsTo notIn mapsTo.

Lemma lookupUpdated (f : A) (c : option C) (m : M) :
  (update f c m) f = c.
  Proof.
    unfold update; rewrite <- beq_a_refl; reflexivity.
  Qed.

Lemma lookupNUpdated (f g : A) (c : option C) (m : M) :
  f <> g -> update f c m g = m g.
  Proof.
    unfold update; remember (beq_a f g) as b; destruct b; intros F.
    - exfalso; apply F. now apply beq_a_true.
    - reflexivity.
  Qed.

Hint Resolve lookupUpdated lookupNUpdated.

Definition singleton (a : A) (c : C) : PartialMap A C := fun x => if beq_a x a then Some c else None.

(* Boring lemmas about partial maps *)

Definition applyNone : option C := None.

Hint Resolve applyNone.

Lemma inDomUpdate : forall m f g c, f <> g -> inDom m g -> inDom (update f c m) g.
  Proof.
    intros m f g c F [y H]; autounfold; rewrite lookupNUpdated; eauto.
  Qed.    

Lemma inDomUpdate' : forall m f g c, f <> g -> inDom (update f c m) g -> inDom m g.
  Proof.
    intros m f g c F [y H]; exists y; rewrite lookupNUpdated in H; eauto.
  Qed.    

Lemma inDomUpdate'' : forall m f c, inDom (update f (Some c) m) f.
  Proof.
    intros; exists c; rewrite lookupUpdated; reflexivity.
  Qed.

Lemma inDomUpdate''' : forall f g m, inDom (update f None m) g -> inDom m g.
  Proof.
    intros f g m [y H]; destruct (dec_eq_a f g); subst.
    - exfalso; rewrite lookupUpdated in *; discriminate.
    - rewrite lookupNUpdated in H; eauto.
  Qed.

Lemma notInUpdate : forall m f g c, f <> g -> notIn m g -> notIn (update f c m) g.
  Proof.
    intros m f g c F; autounfold; destruct (dec_eq_a f g).
      - contradiction; tauto.
      - rewrite lookupNUpdated; eauto.
  Qed.    

Lemma notInUpdate' : forall m f g c, f <> g -> notIn (update f c m) g -> notIn m g.
  Proof.
    autounfold; intros m f g c F H; rewrite lookupNUpdated in H; eauto.
  Qed.    

Lemma notInUpdate'' : forall m f g, notIn m g -> notIn (update f None m) g.
  Proof.
    intros; autounfold; destruct (dec_eq_a f g).
    - subst; now apply lookupUpdated.
    - now rewrite lookupNUpdated.
  Qed.

Hint Resolve inDomUpdate inDomUpdate' inDomUpdate'' inDomUpdate'''
  notInUpdate notInUpdate' notInUpdate''.

Lemma elemContradiction : forall m f, inDom m f -> notIn m f -> False.
  Proof.
    autounfold; intros m f [y E] N; rewrite E in N; discriminate.
  Qed.

Hint Resolve elemContradiction.

(** Separation logic predicates *)

Definition disjoint (m1 m2 : M) : Prop :=
  forall (f : A), (inDom m1 f -> notIn m2 f)
    /\ (inDom m2 f -> notIn m1 f).

Definition union (m1 m2 : M) (H : disjoint m1 m2) : M :=
  fun x =>
    match m1 x with
        | None => m2 x
        | Some y => Some y
    end.

Definition splits (m1 m2 m : M) :=
  { H : disjoint m1 m2 & forall f, union m1 m2 H f = m f}.

(* Boring lemmas about disjointness and splitting *)

Lemma notInSplit : forall m m1 m2 f, splits m1 m2 m -> notIn m1 f -> notIn m2 f ->  notIn m f.
  Proof.
    autounfold; intros m1 m2 m f [H3 H4] H1 H2; rewrite <- H4. unfold union. rewrite H1, H2; reflexivity.
  Qed.

Lemma notInSplitL : forall m m1 m2 f, splits m1 m2 m -> notIn m f -> notIn m1 f.
  Proof.
    intros m m1 m2 f.
    autounfold; unfold splits, union.
    intros [H1 H2] H3; set (H4 := H2 f).
    rewrite H3 in H4.
    destruct (m1 f); [discriminate | trivial].
  Qed.    

Lemma notInSplitR : forall m m1 m2 f, splits m1 m2 m -> notIn m f -> notIn m2 f.
  Proof.
    intros m m1 m2 f.
    autounfold; unfold splits, union.
    intros [H1 H2] H3; set (H4 := H2 f).
    rewrite H3 in H4.
    destruct (m1 f); [discriminate | trivial].
  Qed.    

Hint Resolve notInSplit notInSplitL notInSplitR.

Lemma updateSeparate : forall f c m1 m2, notIn m2 f -> disjoint (update f c m1) m2 -> disjoint m1 m2.
  Proof.
  intros f c m1 m2 N D g; destruct (dec_eq_a f g) as [E | F]; [subst f | ];
    destruct (D g); split; eauto.
  Qed.

Hint Resolve updateSeparate.

Lemma disjointUpdate (m1 m2 m : M) (f : A) (c : option C) :
  disjoint m1 m2 -> notIn m2 f -> disjoint (update f c m1) m2.
  Proof.
    intros D N g; destruct (dec_eq_a f g) as [E | F]; destruct (D g); [ subst f | ]; split; eauto.
  Qed.

Hint Resolve disjointUpdate.

Lemma lookupUnion (m1 m2 m : M) (f : A) (c : option C) :
   notIn m2 f ->
   (forall H f, union m1 m2 H f = m f) ->
   forall H g, (union (update f c m1) m2 H g = update f c m g).
  Proof.
      intros N U H g; destruct (dec_eq_a f g); [subst f | ].
      - unfold union; rewrite lookupUpdated, lookupUpdated; destruct c; trivial.
      - unfold union in *; rewrite lookupNUpdated; [ | assumption]. 
        rewrite U; [ now rewrite lookupNUpdated  | now eauto].        
  Qed.

Hint Resolve lookupUnion.

Lemma splitUpdate (m1 m2 m : M) (f : A) (c : option C) :
  notIn m2 f ->
  splits m1 m2 m -> splits (update f c m1) m2 (update f c m).
  Proof.
    intros N [H1 H2]; assert (H : disjoint (update f c m1) m2) by now auto.
    exists H; eauto.
  Qed.

Hint Resolve splitUpdate.

Lemma singletonMaps : forall a c, mapsTo (singleton a c) a c.
  intros; autounfold; unfold singleton; simpl; rewrite <- beq_a_refl; reflexivity.
  Qed.

Hint Resolve singletonMaps.

Lemma singletonMapsNot : forall a a' c, a' <> a -> notIn (singleton a c) a'.
  intros; unfold singleton; autounfold; simpl.
  remember (beq_a a' a) as b; destruct b.
  - exfalso; now apply H, beq_a_true.
  - reflexivity.    
  Qed.

Hint Resolve singletonMapsNot.

Lemma inSingleton : forall a a' c, inDom (singleton a c) a' -> a = a'.
  unfold singleton; intros a a' c [H1 H2]. 
  remember (beq_a a' a) as b; destruct b; [symmetry; now apply beq_a_true | discriminate ].
  Qed.

Hint Resolve inSingleton.

Lemma notInSingleton : forall m a b c, notIn m a -> inDom m b -> notIn (singleton a c) b.
  intros m a b c H1 H2.
  unfold singleton, notIn.
  remember (beq_a b a) as d; destruct d; [ | reflexivity].
  assert (E : b = a) by now (apply beq_a_true).
  subst; exfalso; eauto.
  Qed.

Hint Resolve notInSingleton.

Lemma singletonSplits : forall a c m, notIn m a -> splits (singleton a c) m (update a (Some c) m).
  Proof.
    intros a c m N; assert (disjoint (singleton a c) m). 
    - intros f; split; eauto.
      intros H; assert (a = f) by apply (inSingleton _ _ _ H); subst; eauto.
    - exists H; intros f.
      unfold union. unfold singleton.
      remember (beq_a f a) as d; destruct d.
      * assert (f = a); [now (apply beq_a_true)| subst; eauto].
      * assert (F : f <> a); [apply beq_a_false; now symmetry | ].
        rewrite lookupNUpdated; auto.
  Qed.

Hint Resolve singletonSplits.

(** Predicates **)

Inductive Predicate : Type :=
  | MAPSTO : A -> C -> Predicate
  | EMPTY: Predicate
  | STAR : Predicate -> Predicate -> Predicate.

Fixpoint holds (p : Predicate) (m : M) : Prop :=
  match p with
    | STAR p1 p2 => exists m1 m2, splits m1 m2 m /\ holds p1 m1 /\ holds p2 m2 
    | MAPSTO f c => mapsTo m f c /\ forall f', f <> f' -> notIn m f'
    | EMPTY => forall f, notIn m f
  end.

Hint Unfold holds.

Notation "lb --> la" := (MAPSTO (next lb) (ptr la)) (at level 90, right associativity).

Notation "l |-> c" := (MAPSTO (contents l) c) (at level 90, right associativity).

(** Commands with their denotational semantics **)

Inductive Cmd : Type :=
  | Insert : L -> L -> L -> Cmd
  | Remove : L -> L -> L -> Cmd
  | Modify : L -> C -> C -> Cmd
  | Seq : Cmd -> Cmd -> Cmd.

Fixpoint semantics (cmd : Cmd) (m : M) : M :=
  match cmd with
   | Insert lb l la => (update (contents l) (Some epsilon) 
                       (update (next l) (Some (ptr la)) 
                       (update (next lb) (Some (ptr l)) m)))
   | Remove lb l la => update (next lb) (Some (ptr la))
                       (update (next l) None
                       (update (contents l) None m))
   | Modify l _ d => update (contents l) (Some d) m
   | Seq c1 c2 => semantics c2 (semantics c1 m)
  end.

Hint Unfold semantics disjoint inDom union.

Fixpoint post (cmd : Cmd) : M -> Prop :=
  match cmd with
    | Insert lb l la => holds (STAR (l |-> epsilon) (STAR (l --> la) (lb --> l)))
    | Remove lb l la => holds (lb --> la)
    | Modify l _ d => holds (l |-> d)
    | Seq c1 c2 => post c2
  end.

Fixpoint pre (cmd : Cmd) : M -> Prop :=
  match cmd with
    | Insert lb l la => fun m => holds (lb --> la) m
                        /\ lb <> l /\ la <> l /\ lb <> la
    | Remove lb l la => fun m => holds (STAR (lb --> l) (STAR (l --> la) (l |-> epsilon))) m
                        /\ lb <> l /\ la <> l /\ lb <> la
    | Modify l c d => holds (l |-> c)
    | Seq c1 c2 => fun m => pre c1 m /\ (forall m, post c1 m -> pre c2 m)
  end.

(* Soundness of semantics *)

Theorem soundness (cmd : Cmd) : forall m, pre cmd m -> post cmd (semantics cmd m).
  Proof.
    induction cmd.
    - simpl; intros m [H1 H2]; destruct_conjs.
      exists (singleton (contents l0) epsilon).
      exists (update (next l0) (Some (ptr l1))
             (update (next l) (Some (ptr l0)) m)).
      repeat (split; eauto).
      exists (singleton (next l0) (ptr l1)).
      exists (update (next l) (Some (ptr l0)) m).
      repeat (split; try apply singletonSplits; eauto).
    - simpl pre; simpl post; intros; destruct_conjs;split; [ now eauto | ]; intros.
      destruct (dec_eq_a (next l0) f'); destruct (dec_eq_a (contents l0) f'); subst; eauto.
        eapply notInUpdate, notInUpdate'', notInUpdate'', notInSplit; eauto.
    - simpl in *; autounfold; intros; split; intros; destruct_conjs; eauto.
      rewrite lookupNUpdated; eauto.
    - autounfold; simpl; intros; destruct_conjs; eauto.
  Qed.

(* Defining mod, addr and separation *)

Inductive Modifies : A -> Cmd -> Prop :=
  | ModInsert1 : forall lb l la, Modifies (contents l) (Insert lb l la)
  | ModInsert2 : forall lb l la, Modifies (next l) (Insert lb l la)
  | ModInsert3 : forall lb l la, Modifies (next lb) (Insert lb l la)
  | ModRemove1 : forall lb l la, Modifies (next lb) (Remove lb l la) 
  | ModRemove2 : forall lb l la, Modifies (next l) (Remove lb l la) 
  | ModRemove3 : forall lb l la, Modifies (contents l) (Remove lb l la) 
  | ModModify : forall l c d, Modifies (contents l) (Modify l c d)
  | ModSeq1 : forall f cmd1 cmd2, Modifies f cmd1 -> Modifies f (Seq cmd1 cmd2) 
  | ModSeq2 : forall f cmd1 cmd2, Modifies f cmd2 -> Modifies f (Seq cmd1 cmd2).

Hint Constructors Modifies.

Inductive Addr : A -> Predicate -> Prop :=
  | DomMaps : forall f c, Addr f (MAPSTO f c)
  | DomStarL : forall f P Q, Addr f P -> Addr f (STAR P Q)
  | DomStarR : forall f P Q, Addr f Q -> Addr f (STAR P Q).

Hint Constructors Addr.

Definition Separate (cmd : Cmd) (R : Predicate) : Prop :=
 forall f, Modifies f cmd -> Addr f R -> False.

Hint Unfold Separate.

Lemma SeparateStarL : forall cmd R1 R2, Separate cmd (STAR R1 R2) -> Separate cmd R1.
  Proof.
    intros; induction cmd; intros; eauto.
  Qed.

Lemma SeparateStarR : forall cmd R1 R2, Separate cmd (STAR R1 R2) -> Separate cmd R2.
  Proof.
    intros; induction cmd; intros; eauto.
  Qed.

Hint Resolve SeparateStarL SeparateStarR.



(* Here is the interesting part of the frame rule proof *)

Lemma splitSemantics (cmd : Cmd) :
     forall m1 m2 m R,
     (forall f, Modifies f cmd -> notIn m2 f) ->
     holds R m2 -> 
     splits m1 m2 m ->
     splits (semantics cmd m1) m2 (semantics cmd m).
  Proof.
    induction cmd; intros; autounfold; eauto; try apply splitUpdate; eauto.
    eapply IHcmd2; eauto.
  Qed.

Hint Resolve splitSemantics.

Lemma SeparateLemma (cmd : Cmd) (R : Predicate) : 
  Separate cmd R ->
  (forall a m, Modifies a cmd -> holds R m -> notIn m a).
  Proof.
    induction R; intros D a' m C H.
     - simpl holds in *; destruct_conjs; destruct (dec_eq_a a a') as [E | F].
       * subst; exfalso; eauto.
       * eauto.
     - eauto.
     - destruct H as [m1 [m2 H]]; destruct_conjs; eapply (notInSplit m m1 m2 _ H); eauto.
  Qed.

Hint Resolve SeparateLemma.
    
Theorem frameRule (cmd : Cmd) (P Q R : Predicate)
        (H : forall m, holds P m -> holds Q (semantics cmd m)) 
        (D : Separate cmd R)
  : forall m, holds (STAR P R) m -> holds (STAR Q R) (semantics cmd m).
  Proof.
    intros m [m1 [m2 A]]; destruct_conjs; exists (semantics cmd m1); exists m2; split; eauto.
  Qed.
